import PySimpleGUI as sg
from leeselab_project_creator import generate_plate_layout, generate_worklist
from pathlib import Path


# main function to call the GUI
def main():
    # define the layout of the GUI
    layout = [
        [
            sg.Text(
                "This is the Leeselab laboratory planing assistant.\nPlease fill the fields to get started."
            )
        ],
        [sg.Text("Path to the list of samples:")],
        [
            sg.InputText(
                size=(40, 1),
                do_not_clear=True,
                key="input_file",
                tooltip="Please select a list of samples in Excel format.",
            ),
            sg.FileBrowse(
                file_types=(("Excel", ".xlsx"),),
                tooltip="Please select a list of samples in Excel format.",
            ),
        ],
        [sg.Text("Excel sheet name:")],
        [
            sg.InputText(
                default_text="Tabelle1",
                key="sheet_name",
                size=(40, 1),
                tooltip="Please write the name of the correct sheet here.",
            )
        ],
        [sg.Text("First sample can be found in:")],
        [
            [
                sg.Text("Column"),
                sg.InputText(
                    size=(10, 1),
                    tooltip="Indicate the column of the first sample (e.g. A)",
                    key="column",
                ),
                sg.Text("Row"),
                sg.InputText(
                    size=(10, 1),
                    tooltip="Indicate the row of the first samples (e.g. 1)",
                    key="row",
                ),
            ]
        ],
        [sg.Text("Number of extraction replicates:")],
        [
            sg.Spin(
                values=list(range(1, 501)),
                initial_value=1,
                size=(5, 1),
                key="extraction_replicates",
                tooltip="How many extraction replicates do you need?",
            )
        ],
        [sg.Text("Number of PCR replicates:")],
        [
            sg.Spin(
                values=list(range(1, 501)),
                initial_value=1,
                size=(5, 1),
                key="pcr_replicates",
                tooltip="How many PCR replicates do you need?",
            )
        ],
        [sg.Text("Primers to use:")],
        [
            sg.InputText(
                size=(40, 1),
                tooltip="Please add all primers you want to use seperated by comma (e.g. fwh2, rbcl)",
                key="markers",
            )
        ],
        [sg.Text("Available 1st step tags:")],
        [
            sg.InputText(
                size=(40, 1),
                tooltip="Please add the number of the available first step tags (e.g. 1, 2, 5, 6)",
                key="primers",
            )
        ],
        [sg.Checkbox(text="Suggest pooling strategy", default=True, key="pooling")],
        [sg.Text("Project name")],
        [
            sg.InputText(
                size=(40, 1),
                tooltip="Please enter a project name here",
                key="project_name",
            )
        ],
        [sg.Text("Select an output folder")],
        [
            sg.InputText(
                size=(40, 1),
                tooltip="Please select an output folder to save the generated sheets",
                key="output_folder",
            ),
            sg.FolderBrowse(
                tooltip="Please select an output folder to save the generated sheets"
            ),
        ],
        [
            sg.Button(
                "Generate sample list",
                button_color="red",
                size=(42, 2),
                disabled=True,
                key="generate",
            )
        ],
        [
            sg.Multiline(
                size=(45, 5),
                key="output_text",
                autoscroll=True,
                default_text="Please fill all fields to generate a project.\n",
            )
        ],
        [
            sg.Button(
                "Cancel",
            )
        ],
    ]

    # define the window to be called
    window = sg.Window("Leeselab project creator", layout)
    # was the output already updated?
    updated_out = False
    # run the main loop
    while True:
        # read the windows values and events every 100 ms
        event, values = window.read(timeout=200)

        # remove those values since they are not needed, easier to check to update the button
        try:
            del values["Browse"]
            del values["Browse0"]
        except TypeError:
            pass

        # if the close button is pressed, break the loop, close the window
        if event == sg.WIN_CLOSED or event == "Cancel":
            break

        # enable generate button when all fields are filled
        if all(values.values()):
            window["generate"].update(disabled=False)
            if not updated_out:
                window["output_text"].print("Ready to generate project files.")
                updated_out = True
            else:
                updated_out = True
        else:
            window["generate"].update(disabled=True)
            updated_out = False

        if event == "generate":
            # generate the plate layout
            generate_plate_layout.generate_plate_layout(
                input_file=Path(values["input_file"]),
                sheet=values["sheet_name"],
                col=values["column"],
                row=values["row"],
                extraction_replicates=values["extraction_replicates"],
                output_folder=Path(values["output_folder"]),
                project_name=values["project_name"],
            )

            # generate the worklist
            generate_worklist.generate_worklist(
                output_path=Path(values["output_folder"]),
                project=values["project_name"],
                available_primers=values["primers"],
                pcr_replicates=values["pcr_replicates"],
                markers=values["markers"],
            )
            window["output_text"].print(
                "Plate layout and worklist saved to output folder."
            )

    window.Close()


# run only if called as a toplevel script
if __name__ == "__main__":
    main()



import openpyxl
import pandas as pd
from openpyxl.styles import Alignment, PatternFill, Font
from openpyxl.styles.borders import Border, Side


# funtion to calculate one worklist per marker
def worklist_per_marker(
    available_primers,
    extraction_plates,
    pcr_replicates,
    primer_name,
    optimal_primers,
    starting_library_number,
):
    # generate the output dataframe
    output_worklist = pd.DataFrame()
    output_worklist["source plate"] = [
        extraction_plate
        for extraction_plate in extraction_plates
        for _ in range(pcr_replicates)
    ]

    # find the number of pcr plates for this marker
    pcr_plate_count = len(output_worklist)

    # find the number of librarys to distribute primers evenly
    available_primers = available_primers.split(",")
    available_primers = [int(primer) for primer in available_primers]
    number_of_primers = len(available_primers)

    while True:
        if pcr_plate_count <= number_of_primers:
            library_count = 1
            break
        elif pcr_plate_count % number_of_primers == 0:
            library_count = int(pcr_plate_count / number_of_primers)
            break
        else:
            number_of_primers -= 1

    # calculate the number of plates per library
    plates_per_library = int(pcr_plate_count / library_count)

    # generate the library column
    library_column = []

    for i in range(library_count):
        for _ in range(plates_per_library):
            library_column.append(starting_library_number)
        starting_library_number += 1

    output_worklist["library"] = library_column

    primers = []

    # find the optimal primers for maximum library diversity
    for primer in optimal_primers:
        if len(primers) < plates_per_library:
            if primer not in primers and primer in available_primers:
                primers.append(primer)
        else:
            break

    primers = primers * library_count
    primers = [
        "{} - {}".format(primer_name, primer_number) for primer_number in primers
    ]

    output_worklist["tagging primer"] = primers
    output_worklist["1st pcr"] = ""
    output_worklist["clean up"] = ""
    output_worklist["2nd pcr"] = ""
    output_worklist["normalization"] = ""
    output_worklist["pooling"] = ""

    return (
        output_worklist,
        starting_library_number,
        plates_per_library,
    )


def generate_worklist(output_path, project, available_primers, pcr_replicates, markers):
    # generate the correct path to the file
    input_file = "{}_plate_layout.xlsx".format(project)
    input_file = output_path.joinpath(input_file)

    # read the data
    data = pd.read_excel(input_file, sheet_name="plate_layout", skiprows=1)

    # collect all worklists here for the final output
    worklists = []

    # collect the plate letters
    plates = list(data["Lysis,\nExtraction,\nPCR plate"].unique())
    general_worklist = pd.DataFrame(
        columns=[
            "plate",
            "aliquoted",
            "lysis",
            "inhibitor removal",
            "lysate distribution",
            "extraction",
            "gel extraction",
        ]
    )

    # general worklist is done
    general_worklist["plate"] = plates

    # optimal primer order for maximizing library diversity
    optimal_primer_order = [
        1,
        5,
        9,
        13,
        17,
        21,
        2,
        6,
        10,
        14,
        18,
        22,
        3,
        7,
        11,
        15,
        19,
        23,
        4,
        8,
        12,
        16,
        20,
        24,
    ]

    # gather the worklists here
    worklists = []

    # calculate everything marker wise and concat the marker dfs in the end
    next_library = 1
    markers = markers.split(",")

    for primer in markers:
        worklist, next_library, plates_per_library = worklist_per_marker(
            available_primers,
            plates,
            pcr_replicates,
            primer,
            optimal_primer_order,
            next_library,
        )

        worklists.append(worklist)

    # concat the individual worklist to have a working table
    working_table = pd.concat(worklists, axis=0)

    # save both tables to excel to perform styling via openpyxl
    # tables will be saved as sheets to the plate layout
    with pd.ExcelWriter(
        input_file, mode="a", if_sheet_exists="replace", engine="openpyxl"
    ) as writer:
        general_worklist.to_excel(writer, sheet_name="extraction_worklist", index=False)
        working_table.to_excel(writer, sheet_name="pcr_worklist", index=False)

    ## add the styling
    add_styling(input_file, project, plates_per_library)


# function to add styling to the worklists
def add_styling(input_file, project, plates_per_library):

    # open the extraction worklist first
    wb = openpyxl.load_workbook(input_file)
    ws = wb["extraction_worklist"]

    # Iterate over all columns and adjust their widths
    for column in ws.columns:
        max_length = 0
        column_letter = column[0].column_letter
        for cell in column:
            try:
                if len(str(cell.value)) > max_length:
                    max_length = len(cell.value)
            except:
                pass
        adjusted_width = (max_length + 2) * 1.2
        ws.column_dimensions[column_letter].width = adjusted_width

    # add a connected header cell
    ws.insert_rows(1)
    ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=7)
    ws.cell(row=1, column=1).value = "Extraction worklist: {}".format(project)

    # styling for the header
    ws.cell(row=1, column=1).font = Font(bold=True)
    ws.cell(row=1, column=1).alignment = Alignment(horizontal="center")

    # add alternatig colors
    ## add alternating cell styling and borders
    thin_border = Border(
        left=Side(style="thin"),
        right=Side(style="thin"),
        top=Side(style="thin"),
        bottom=Side(style="thin"),
    )

    fill = PatternFill(start_color="00C0C0C0", end_color="00C0C0C0", fill_type="solid")

    for row in range(2, ws.max_row + 1):
        for col in range(1, 8):
            ws.cell(row=row, column=col).border = thin_border
            if row % 2 == 0:
                ws.cell(row=row, column=col).fill = fill

    # merge cells where steps are executed together
    for col in range(2, 6):
        for row in range(3, ws.max_row + 1, 2):
            ws.merge_cells(
                start_row=row, end_row=row + 1, start_column=col, end_column=col
            )

    # add styling for the pcr worklist
    ws = wb["pcr_worklist"]

    # Iterate over all columns and adjust their widths
    for column in ws.columns:
        max_length = 0
        column_letter = column[0].column_letter
        for cell in column:
            try:
                if len(str(cell.value)) > max_length:
                    max_length = len(cell.value)
            except:
                pass
        adjusted_width = (max_length + 2) * 1.2
        ws.column_dimensions[column_letter].width = adjusted_width

    # add a connected header cell
    ws.insert_rows(1)
    ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=8)
    ws.cell(row=1, column=1).value = "PCR worklist: {}".format(project)

    # styling for the header
    ws.cell(row=1, column=1).font = Font(bold=True)
    ws.cell(row=1, column=1).alignment = Alignment(horizontal="center")

    # add alternatig colors
    ## add alternating cell styling and borders
    thin_border = Border(
        left=Side(style="thin"),
        right=Side(style="thin"),
        top=Side(style="thin"),
        bottom=Side(style="thin"),
    )

    fill = PatternFill(start_color="00C0C0C0", end_color="00C0C0C0", fill_type="solid")

    for row in range(2, ws.max_row + 1):
        for col in range(1, 9):
            ws.cell(row=row, column=col).border = thin_border
            if row % 2 == 0:
                ws.cell(row=row, column=col).fill = fill

    # merge cells where steps are executed together
    for col in range(7, 9):
        for row in range(3, ws.max_row + 1, plates_per_library):
            ws.merge_cells(
                start_row=row,
                end_row=row + plates_per_library - 1,
                start_column=col,
                end_column=col,
            )

    wb.save(input_file)
    wb.close()



import pandas as pd
import openpyxl
from string import ascii_uppercase
from openpyxl.styles import Alignment, PatternFill, Font
from openpyxl.worksheet.dimensions import ColumnDimension, DimensionHolder
from openpyxl.utils import get_column_letter
from openpyxl.styles.borders import Border, Side


# function to generate the plate layout
def generate_plate_layout(
    input_file, sheet, col, row, extraction_replicates, output_folder, project_name
):
    # open the excel file with the sheetname
    wb = openpyxl.load_workbook(input_file, data_only=True)

    # open the worksheet
    ws = wb[sheet]

    # convert the inputs to the correct format
    letter_to_numeric = list(ascii_uppercase) + [
        i + j for i in ascii_uppercase for j in ascii_uppercase
    ]
    letter_to_numeric = dict(
        zip(letter_to_numeric, range(1, len(letter_to_numeric) + 2))
    )

    col = col.capitalize()
    col = letter_to_numeric[col]
    row = int(row)

    # extract the samples
    samples = [ws.cell(row=row, column=col).value for row in range(row, ws.max_row + 1)]
    sample_count = len(samples)
    replicate_numbers = sample_count * list(range(1, extraction_replicates + 1))

    # include the extraction replicates
    samples = sorted(list(enumerate(samples)) * extraction_replicates)
    samples = [
        "{}_{}".format(sample[1], number)
        for sample, number in zip(samples, replicate_numbers)
    ]

    ## first sample will add with every step in the while loop
    number = 1
    letter = 0
    well = 1
    ncs = [8, 9, 21, 27, 39, 44, 54, 58, 68, 73, 88, 93]
    nc = 1
    output_dict = {}
    letters = list(ascii_uppercase) + [
        i + j for i in ascii_uppercase for j in ascii_uppercase
    ]

    # add a dict with alphanumeric convention
    alpha = [
        "{}{}".format(letter, well)
        for well in range(1, 13)
        for letter in ["A", "B", "C", "D", "E", "F", "G", "H"]
    ]
    alpha = dict(zip(range(1, 97), alpha))

    while samples:
        ## if the plate and the replicate is full jump to plate 3
        if (number - 1) % 96 == 0 and number != 1:
            number += 96
            letter += 2
            well = 1
            nc = 1

        ## if well contains a negative controll, add negative control to the output dict
        if well in ncs:
            ## first replicate
            output_dict[number] = [
                "NC_{}_{}{}".format(nc, letters[letter], letters[letter + 1]),
                letters[letter],
                alpha[well],
            ]
            output_dict[number + 96] = [
                "NC_{}_{}{}".format(nc, letters[letter], letters[letter + 1]),
                letters[letter + 1],
                alpha[well],
            ]
            nc += 1
            number += 1
            well += 1
        ## else add a sample
        else:
            sample = samples.pop(0)
            output_dict[number] = ["{}".format(sample), letters[letter], alpha[well]]
            output_dict[number + 96] = [
                "{}".format(sample),
                letters[letter + 1],
                alpha[well],
            ]
            number += 1
            well += 1

    out_df = []

    for key in sorted(output_dict.keys()):
        out_df.append(output_dict[key])

    out_df = pd.DataFrame(
        out_df,
        columns=[
            "Sample name",
            "Lysis,\nExtraction,\nPCR plate",
            "Lysis,\nExtraction,\nPCR Well",
        ],
    )

    # order for tubes
    tube_order = [
        "{}{}".format(letter, number)
        for letter in ["A", "B", "C", "D", "E", "F"]
        for number in reversed(range(1, 9))
    ]

    alpha = [
        "{}{}".format(letter, well)
        for well in range(1, 13)
        for letter in ["A", "B", "C", "D", "E", "F", "G", "H"]
    ]

    # add tube order
    tube_order = dict(zip(alpha, tube_order + tube_order))
    out_df["Lysis\ntube"] = out_df["Lysis,\nExtraction,\nPCR Well"].map(tube_order)

    # add tagging primers
    alpha = dict(zip(alpha, range(1, 97)))
    out_df["Aliquotiert"] = ""
    out_df["Primer\n2nd PCR"] = out_df["Lysis,\nExtraction,\nPCR Well"].map(alpha)
    out_df["Notes"] = ""
    out_df = out_df.reset_index(names="#")
    out_df["#"] = out_df["#"] + 1

    # create a temporary save before doing some styling
    savename = "{}_plate_layout.xlsx".format(project_name)
    savename = output_folder.joinpath(savename)
    out_df.to_excel(savename, index=False, sheet_name="plate_layout")
    wb.close()

    # use openpyxl to style everything
    wb = openpyxl.load_workbook(savename)
    ws = wb["plate_layout"]

    # add the line breaks for styling
    header = ["A1", "B1", "C1", "D1", "E1", "F1", "G1", "H1"]
    for cell in header:
        ws[cell].alignment = Alignment(wrap_text=True)

    # adjust cell width for all columns
    dim_holder = DimensionHolder(worksheet=ws)
    widths = [5, 25, 12, 12, 8, 12, 10, 25]

    for col, width in zip(range(ws.min_column, ws.max_column + 1), widths):
        if col == 2:
            dim_holder[get_column_letter(col)] = ColumnDimension(
                ws, min=col, max=col, width=width
            )
        else:
            dim_holder[get_column_letter(col)] = ColumnDimension(
                ws, min=col, max=col, width=width
            )

    ws.column_dimensions = dim_holder

    ## add a header
    ws.insert_rows(1)
    ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=8)
    ws.cell(row=1, column=1).value = "Plate Layout Project: {}".format(project_name)

    ## add alternating cell styling and borders
    thin_border = Border(
        left=Side(style="thin"),
        right=Side(style="thin"),
        top=Side(style="thin"),
        bottom=Side(style="thin"),
    )

    fill = PatternFill(start_color="00C0C0C0", end_color="00C0C0C0", fill_type="solid")

    for row in range(2, ws.max_row + 1):
        for col in range(1, 9):
            ws.cell(row=row, column=col).border = thin_border
            if row % 2 != 0:
                ws.cell(row=row, column=col).fill = fill

    # styling for the header
    ws.cell(row=1, column=1).font = Font(bold=True)
    ws.cell(row=1, column=1).alignment = Alignment(horizontal="center")

    wb.save(savename)
    wb.close()
